#include <iostream>
#include <fstream>
#include "mdfrich.h"

void decodeRICHx80( std::fstream *, rawHeader_t & );
void decodeRICHx81( std::fstream *, rawHeader_t & );
void decodeRICHxC0( std::fstream *, rawHeader_t & );
void decodeODIN( std::fstream *, unsigned );
void decodeOther( std::fstream *, unsigned );
void decodeLHCbNZS( std::fstream * );
void decodeAliceNZS( std::fstream * );
void decodeLHCbZS( std::fstream *, unsigned );
void decodeAliceZS( std::fstream *, unsigned );

unsigned logLevel = 2;
unsigned start = 0;
unsigned seen = 0;
bool dumpHPD = false;
bool mdfCheck = true;
unsigned icount = 0;

int main(int argc, char **argv)
{
  mdfHeader_t mdfHeader;
  eventHeader_t eventHeader;
  rawHeader_t rawHeader;

  for ( unsigned arg=0;arg<argc;arg++ )
  {
    if ( !strcmp( argv[arg], "--logLevel" ) )
    {
      logLevel = strtoul( argv[arg+1], 0, 0 );
      arg++;
    }
    else if ( !strcmp( argv[arg], "--start" ) )
    {
      start = strtoul( argv[arg+1], 0, 0 );
      arg++;
    }
    else if ( !strcmp( argv[arg], "--dont-check-mdf" ) )
    {
      mdfCheck = false;
    }
  }

  std::fstream *fs = new std::fstream(argv[argc-1],std::ios::in|std::ios::binary);
  std::streampos pos = 0;

  for ( ;; )
  {
    fs->seekg( pos );
    fs->read( (char *)&mdfHeader, sizeof(mdfHeader_t) );
    if ( !(*fs) ) break;

    if ( mdfCheck && (mdfHeader.version != 3) )
    {
      printf( "Not an MDF type 3 header. Exit.\n" );
      break;
    }

    pos += mdfHeader.length;

    seen++;
    icount++;
    if ( seen <= start ) continue;

    if ( logLevel > 0 )
    {
      printf( "\nMDF: Header type=%d, Header size=%#x, Data length=%#x\n", mdfHeader.version, mdfHeader.size, mdfHeader.length );
    }

    fs->read( (char *)&eventHeader, sizeof(eventHeader_t) );

    for ( unsigned count=sizeof(mdfHeader_t)+sizeof(eventHeader_t); count<mdfHeader.length;count+=rawHeader.length )
    {
      fs->read( (char *)&rawHeader, sizeof(rawHeader_t) );

      if ( rawHeader.length < 8 ) break;

      switch ( rawHeader.type )
      {
      case 0x9:
        if ( logLevel > 1 ) printf( "  RICH(%#04x) Length=%#6x, SourceId=%#6x\n", rawHeader.version, rawHeader.length, rawHeader.source );
        if ( rawHeader.version == 0x81 )
	{
          decodeRICHx81( fs, rawHeader );
	}
        else if ( rawHeader.version == 0x80 )
	{
          decodeRICHx80( fs, rawHeader );
	}
        else if ( rawHeader.version == 0xC0 )
	{
          decodeRICHxC0( fs, rawHeader );
	  //          pos += 4;
	}
	break;
      case 0x10:
	if ( logLevel > 1 ) printf( "  ODIN(%#04x) Length=%#6x, SourceId=%#6x\n", rawHeader.version, rawHeader.length, rawHeader.source );
        decodeODIN( fs, rawHeader.length-8 );
        break;
      default:
	if ( logLevel > 1 ) printf( "  Other(%#04x) Length=%#6x, SourceId=%#6x\n", rawHeader.version, rawHeader.length, rawHeader.source );
        decodeOther( fs, rawHeader.length-8 );
        break;
      }

    }
  }

  fs->close();

}

// TB2012

void decodeRICHxC0( std::fstream *fs, rawHeader_t &rh )
{
  ingressHeader_t ingressHeader;
  l1Header_xc0_t l1Header;

  unsigned length = rh.length - 8;

  for ( unsigned count=0;count<length; )
  {
    fs->read( (char *)&ingressHeader, sizeof(ingressHeader_t) );
    count += sizeof(ingressHeader_t);
    if ( logLevel > 1 ) printf( "    Ingress[%d]: Channels=%#05x, EventID=%#04x, BXID=%#04x, Truncated=%d\n", ingressHeader.id, ingressHeader.channelMask, ingressHeader.event, ingressHeader.bxid, ingressHeader.truncated );

    if ( ingressHeader.truncated == 1 ) continue;

    unsigned mask = ingressHeader.channelMask;

    for ( unsigned ch=0;ch<12;ch++ )
    {

      if ( (mask & 0x1) == 0x1 )
      {
        fs->read( (char *)&l1Header, sizeof(l1Header_xc0_t) );
        count += sizeof(l1Header_xc0_t);

        if ( logLevel > 2 ) printf( "      [%d,%d,%d,%#05x(%04d)] [G,F,M,Z]=[%1d,%1d,%1d,%1d] NZ count=%#x\n", rh.source, ingressHeader.id, ch, l1Header.hpd, l1Header.hpd, l1Header.gt, l1Header.ext, l1Header.mode, l1Header.zs, l1Header.nzcount );

	//        printf( "[%d] Timestamp:%08x\n", icount, l1Header.timestamp );
        if ( logLevel > 2 ) printf( "      [%d,%#05x] Event=%#04x, Frame=%#05x, Timestamp=%#06x\n", ch, l1Header.hpd, l1Header.evid, l1Header.frame, l1Header.timestamp );

	unsigned char pddata[32];
        fs->read( (char *)pddata, 32 );
        count += 32;

	//        if (logLevel >3)
	{
          printf("[%d] ",icount);
	  for (unsigned b=0;b<32;b+=4)
	  {
            printf("%02x%02x",pddata[b],pddata[b+1]);
	  }
          printf("\n");
	}

      }
      mask = (mask >> 1);
    }
  }
  return;
}

void decodeRICHx81( std::fstream *fs, rawHeader_t &rh )
{
  ingressHeader_t ingressHeader;
  l1Header_x81_t l1Header;
  l0Header_t l0Header;
  l1Trailer_t l1Trailer;

  unsigned length = rh.length - 8;

  for ( unsigned count=0;count<length; )
  {
    fs->read( (char *)&ingressHeader, 4 );
    count += 4;
    if ( logLevel > 1 ) printf( "    Ingress[%d]: Channels=%#05x, EventID=%#04x, BXID=%#04x, Truncated=%d\n", ingressHeader.id, ingressHeader.channelMask, ingressHeader.event, ingressHeader.bxid, ingressHeader.truncated );

    if ( ingressHeader.truncated == 1 ) continue;

    unsigned mask = ingressHeader.channelMask;

    for ( unsigned ch=0;ch<12;ch++ )
    {

      if ( (mask & 0x1) == 0x1 )
      {
        fs->read( (char *)&l1Header, 4 );
        count += 4;
        if ( logLevel > 2 ) printf( "      [%d,%d,%d,%#05x(%04d)] [G,F,M,Z]=[%1d,%1d,%1d,%1d] NZ count=%#x\n", rh.source, ingressHeader.id, ch, l1Header.hpd, l1Header.hpd, l1Header.gt, l1Header.ext, l1Header.mode, l1Header.zs, l1Header.nzcount );
        if (l1Header.gt == 1 )
        {
          mask = (mask >> 1);
          continue;
        }
        if ( l1Header.ext == 1 )
	{
          dumpHPD = true;
          printf( "L1:%08x\n",*(unsigned *)&l1Header );
          fs->read( (char *)&l0Header, 8 );
	  //          printf( "L0:%08x\n",*(unsigned *)&);
          count += 8;
          if ( logLevel > 2 ) printf( "      [%d,%#05x] Event=%#04x, BX=%#05x, [M,T,C]=[%1d,%1d,%1d]\n", ch, l0Header.hpd, l0Header.event, l0Header.bxid, l0Header.mode, l0Header.testPattern, l0Header.calibration );
	}
        if ( l1Header.zs == 0 )
	{
          if (l1Header.mode == 0)
	  {
            decodeLHCbNZS( fs );
            count += 128;
	  }
          else
	  {
            decodeAliceNZS( fs );
            count += 1024;
	  }
	}
        else
	{
          if (l1Header.mode == 0)
	  {
            decodeLHCbZS( fs, l1Header.nzcount );
	  }
          else
	  {
            decodeAliceZS( fs, l1Header.nzcount );
	  }
          count += 4*((l1Header.nzcount + 1)/2);
	}
        if ( l1Header.ext == 1 )
	{
          fs->read( (char *)&l1Trailer, 4 );
          count += 4;
	}
        dumpHPD = false;
      }
      mask = (mask >> 1);
    }
  }
  return;
}

void decodeRICHx80( std::fstream *fs, rawHeader_t &rh )
{
  l1Header_x80_t l1Header;
  l0Header_t l0Header;
  l1Trailer_t l1Trailer;
  unsigned ch = 0;

  unsigned length = rh.length - 8;

  for ( unsigned count=0;count<length; )
  {


    fs->read( (char *)&l1Header, 4 );
    count += 4;

    fs->read( (char *)&l0Header, 8 );
    count += 8;

    if ( logLevel > 2 ) printf( "    [%#05x] [M,Z]=[%1d,%1d] NZ count=%#x, Event [L1,L0]=[%#04x,%#04x], BX=%#05x\n", l0Header.hpd, l1Header.mode, l1Header.zs, l1Header.nzcount, l1Header.event, l0Header.event, l0Header.bxid );

    if ( l1Header.zs == 0 )
    {
      if (l1Header.mode == 0)
      {
        decodeLHCbNZS( fs );
        count += 128;
      }
      else
      {
        decodeAliceNZS( fs );
        count += 1024;
      }
    }
    else
    {
      if (l1Header.mode == 0)
      {
        decodeLHCbZS( fs, l1Header.nzcount - 1 );
      }
      else
      {
        decodeAliceZS( fs, l1Header.nzcount - 1 );
      }
      count += (4*(l1Header.nzcount / 2));
    }

    fs->read( (char *)&l1Trailer, 4 );
    count += 4;

    ch++;
  }

  return;
}

void decodeLHCbNZS( std::fstream *fs )
{
  unsigned row[32];
  fs->read( (char *)row, 128 );
  if ( logLevel > 3 )
  {
    for ( unsigned i=0; i<32; i++ )
    {
      printf( "        %08x\n", row[i] );
    }
  }

  return;
}

void decodeAliceNZS( std::fstream *fs )
{
  unsigned row[256];
  fs->read( (char *)row, 1024 );
  if ( logLevel > 3 )
  {
    for ( unsigned i=0; i<256; i++ )
    {
      printf( "        %08x\n", row[i] );
    }
  }

  return;
}

void decodeLHCbZS( std::fstream *fs, unsigned length )
{
  char tmp[2];
  unsigned length16 = length;
  for ( unsigned i=0; i<length16; i++ )
  {
    fs->read( tmp, 2 );
    if ( (logLevel > 3) && (tmp[0] != 0) )
    {
      printf( "[%02x,%02x] ", tmp[1], tmp[0] );
    }
    if ( dumpHPD )
    {
      printf("%02x%02x\n",tmp[1],tmp[0]);
    }
  }
  
  if ( logLevel > 3 ) printf( "\n" );

  if ( (length & 0x1) == 0x1 )
  {
    fs->read( tmp, 2 );
  }

  return;
}

void decodeAliceZS( std::fstream *fs, unsigned length )
{
  unsigned char pixels[1024];
  unsigned addr;
  unsigned char tmp[2];
  unsigned oddOffset = 0xffffffff;
  unsigned evenOffset = 0xffffffff;
  unsigned length16 = 2*((length + 1)/2);
  for ( unsigned i=0; i<length16; i++ )
  {
    fs->read( (char *)tmp, 2 );
    if ( tmp[1] == 0 ) evenOffset++;
    if ( tmp[1] == 1 ) oddOffset++;
    if ( (tmp[1] & 0x1) == 0 )
    {
      addr = (evenOffset << 8 | tmp[1]) & 0x3ff;
    }
    else
    {
      addr = (oddOffset << 8 | tmp[1]) & 0x3ff;
    }
    pixels[addr] = tmp[0];
  }

  if ( (length16 & 0x1) == 0x1 )
  {
    fs->read( (char *)tmp, 2 );
  }

  if ( logLevel > 3 )
  {
    for ( unsigned i=0; i<1024; i+=4 )
    {
      printf( "        %02x%02x%02x%02x\n", pixels[i+3], pixels[i+2], pixels[i+1], pixels[i] );
    }
  }

  return;
}

void decodeODIN( std::fstream *fs, unsigned length )
{
  odin_t odin;

  fs->read( (char *)&odin, sizeof(odin_t) );
  if ( logLevel > 1 ) printf( "    Event=%#010x%08x, BX=%#05x, GPS=%#010x%08x\n", odin.eventId[0], odin.eventId[1], odin.bunchId, odin.gpsTime[0], odin.gpsTime[1] );

  return;
}

void decodeOther( std::fstream *fs, unsigned length )
{
  char tmp;

  for (unsigned i=0; i<length; i++ )
  {
    fs->read( &tmp, 1 );
  }
  return;
}
